import inspect
from functools import wraps

from sanic import Blueprint, Sanic
from sanic.blueprints import FutureMiddleware


class _Sanic(Sanic):

    def blueprint(self, blueprint, **options):
        super().blueprint(blueprint, **options)
        if hasattr(blueprint, 'nested'):
            for bp in blueprint.nested:
                super().blueprint(bp, **options)


class _Blueprint(Blueprint):

    def __init__(
        self,
        name,
        url_prefix,
        *,
        host=None,
        version=None,
        strict_slashes=False
    ):
        super().__init__(
            name,
            url_prefix=url_prefix,
            host=host,
            version=version,
            strict_slashes=strict_slashes,
        )

        self._blueprints = []

    @property
    def nested(self):
        for bp in self._blueprints:
            yield bp

    def register(self, app, options):
        if 'url_prefix' in options:
            options.pop('url_prefix')
        super().register(app, options)

    def bp(self, name, url_prefix):
        new_url_prefix = self.url_prefix + url_prefix
        new_name = '.'.join([self.name, name])
        bp = _Blueprint(
            new_name,
            url_prefix=new_url_prefix,
            host=self.host,
            version=self.version,
            strict_slashes=self.strict_slashes,
        )
        self._blueprints.append(bp)
        return bp


def bp_middleware(bp, attach_to='request'):

    def middleware(f):

        @wraps(f)
        async def inner_middleware(request, *args):
            if request.path.startswith(bp.url_prefix):
                if attach_to == 'request':
                    response = f(request)
                elif attach_to == 'response':
                    response = f(request, *args)
                else:
                    # TODO see what fits best here, an custom exception or
                    # just return None?
                    raise Exception('or return None?')  # noqa

                if inspect.isawaitable(response):
                    response = await response  # noqa
                return response

        future_middleware = FutureMiddleware(inner_middleware, [attach_to], {})
        bp.middlewares.append(future_middleware)
        return future_middleware

    return middleware


__all__ = ['_Blueprint', '_Sanic', 'bp_middleware']



class SanicToolboxException(Exception):
    pass


class BeforeAndAfterNotSupported(SanicToolboxException):
    pass


class ListenerNotFound(SanicToolboxException):
    pass


class TaskNotFound(SanicToolboxException):
    pass


class MiddlewareNotFound(SanicToolboxException):
    pass



import importlib
import inspect
import logging
import sys
from pathlib import Path

logger = logging.getLogger(__name__)


def search_modules(root, *args):
    if root not in sys.modules:
        raise ValueError('"{}" is not a valid (or loaded) module'.format(root))

    root_mod = sys.modules[root]
    root_mod_path = Path(root_mod.__file__).parent
    for a in args:
        root_mod_path /= a
    if not root_mod_path.is_dir():
        raise ValueError(
            '"{!s}" is not a valid directory'.format(root_mod_path)
        )

    relative_module_root = [root, *args]
    for module_name in sorted(
        map(inspect.getmodulename, root_mod_path.glob('*.py'))
    ):
        if module_name == '__init__':
            continue

        yield [*relative_module_root, module_name]


def load_module(module_name):
    logger.debug('importing "{name}"'.format(name='.'.join(module_name)))
    try:
        return importlib.import_module('.'.join(module_name))

    except ImportError as ie:
        logger.exception(
            'could not import module "{}": {}'.format(
                '.'.join(module_name), ie.msg
            )
        )
        return None


__all__ = ['search_modules', 'load_module']



from collections import defaultdict, deque
from copy import deepcopy
from functools import partial

from .exceptions import (
    BeforeAndAfterNotSupported,
    ListenerNotFound,
    MiddlewareNotFound,
    TaskNotFound,
)


class LazyApplication:
    """Most of the code here was taken directly from Sanic Application class.
    I'm kind of ashame of that, but it'll do for now.
    """

    def __init__(self):
        self._listeners = defaultdict(list)
        self._request_middleware = deque()
        self._response_middleware = deque()
        self._tasks = deque()

    def listener(self, event, *, before=None, after=None):
        """Create a listener from a decorated function.
        :param event: event to listen to
        """

        def decorator(listener):
            if any([before, after]):
                if all([before, after]):
                    raise BeforeAndAfterNotSupported

            if before:
                if before not in self._listeners[event]:
                    raise ListenerNotFound

                idx = self._listeners[event].index(before)
                self._listeners[event].insert(idx, listener)
            elif after:
                if after not in self._listeners[event]:
                    raise ListenerNotFound

                idx = self._listeners[event].index(after)
                self._listeners[event].insert(idx + 1, listener)
            else:
                self._listeners[event].append(listener)
            return listener

        return decorator

    def register_middleware(
        self, middleware, attach_to='request', before=None, after=None
    ):

        attach_middleware_to = []
        if attach_to == 'request':
            attach_middleware_to = self._request_middleware
        if attach_to == 'response':
            attach_middleware_to = self._response_middleware

        if any([before, after]):
            if all([before, after]):
                raise BeforeAndAfterNotSupported

        if before:
            if before not in attach_middleware_to:
                raise MiddlewareNotFound

            idx = attach_middleware_to.index(before)
            attach_middleware_to.insert(idx, middleware)
        elif after:
            if after not in attach_middleware_to:
                raise MiddlewareNotFound

            idx = attach_middleware_to.index(after)
            attach_middleware_to.insert(idx + 1, middleware)
        else:
            attach_middleware_to.append(middleware)
        return middleware

    def middleware(self, middleware_or_request, *, before=None, after=None):
        """Decorate and register middleware to be called before a request.
        Can either be called as @app.middleware or @app.middleware('request')
        """
        # Detect which way this was called, @middleware or @middleware('AT')
        if callable(middleware_or_request):
            return self.register_middleware(
                middleware_or_request, before=before, after=after
            )

        else:
            return partial(
                self.register_middleware,
                attach_to=middleware_or_request,
                before=before,
                after=after,
            )

    def add_task(self, task, *, before=None, after=None):
        """Schedule a task to run later, after the loop has started.
        Different from asyncio.ensure_future in that it does not
        also return a future, and the actual ensure_future call
        is delayed until before server start.
        :param task: future, couroutine or awaitable
        """

        if any([before, after]):
            if all([before, after]):
                raise BeforeAndAfterNotSupported

        if before:
            if before not in self._tasks:
                raise TaskNotFound

            idx = self._tasks.index(before)
            self._tasks.insert(idx, task)
        elif after:
            if after not in self._tasks:
                raise TaskNotFound

            idx = self._tasks.index(after)
            self._tasks.insert(idx + 1, task)
        else:
            self._tasks.append(task)

    def clone(self):
        lazyapp = LazyApplication()
        lazyapp._listeners = deepcopy(self._listeners)
        lazyapp._request_middleware = deepcopy(self._request_middleware)
        lazyapp._response_middleware = deepcopy(self._response_middleware)
        lazyapp._tasks = deepcopy(self._tasks)
        return lazyapp

    def apply(self, app, sanic_response_order, sanic_listener_stop_order):
        for task in self._tasks:
            app.add_task(task())
        for middleware in self._request_middleware:
            app.register_middleware(middleware, 'request')
        if not sanic_response_order:
            self._response_middleware.reverse()
        for middleware in self._response_middleware:
            app.register_middleware(middleware, 'response')
        for event, listeners in self._listeners.items():
            if event in (
                'before_server_stop', 'after_server_stop'
            ) and not sanic_listener_stop_order:
                listeners.reverse()
            for listener in listeners:
                app.listener(event)(listener)

        self.clear()
        return app

    def clear(self):
        # clear everything
        self._tasks.clear()
        self._request_middleware.clear()
        self._response_middleware.clear()
        self._listeners.clear()

    def __call__(
        self, app, sanic_response_order=True, sanic_listener_stop_order=True
    ):
        return self.apply(
            app,
            sanic_response_order=sanic_response_order,
            sanic_listener_stop_order=sanic_listener_stop_order,
        )


def lazyapp():
    return LazyApplication()


__all__ = ('lazyapp')



import logging

from .blueprint import _Blueprint as Blueprint
from .blueprint import _Sanic as Sanic
from .blueprint import bp_middleware
from .lazyapp import lazyapp
from .utils import load_module, search_modules

logging.getLogger(__name__).addHandler(logging.NullHandler())

__version__ = '0.3.0'
__description__ = 'A general utility toolbox for Sanic to make middleware workflows simple and presumable, without monkey patching, for plugins and applications'


__all__ = [
    'lazyapp',
    'Blueprint',
    'Sanic',
    'bp_middleware',
    'search_modules',
    'load_module',
]



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

tests_require = [
    'coverage',
    'pytest-cov',
    'pytest-flakes',
    'pytest-pep8',
    'pytest-asyncio',
    'pytest',
    'sanic',
    'codecov',
]

extras_require = {
    'docs': ['Sphinx'],
    # 'sphinx_rtd_theme',
    'tests': tests_require,
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)

setup_requires = ['pytest-runner']

install_requires = []

about = {'version': '0.3.0', 'description': ''}

with open(
    os.path.join(here, 'sanic_toolbox', '__init__.py'), 'r', encoding='utf-8'
) as f:
    for line in f:
        if line.startswith('__version__'):
            about['version'] = line.strip().split('=')[1].strip(' \'"')
        if line.startswith('__description__'):
            about['description'] = line.strip().split('=')[1].strip(' \'"')

setup(
    name='sanic-toolbox',
    version=about['version'],
    description=about['description'],
    url='https://github.com/vltr/sanic-toolbox',
    download_url='https://github.com/vltr/sanic-toolbox/archive/master.zip',
    author='Richard Kuesters',
    author_email='rkuesters@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='sanic toolbox utils plugin',
    packages=find_packages(exclude=['example', 'tests']),
    install_requires=install_requires,
    extras_require=extras_require,
    setup_requires=setup_requires,
    tests_require=tests_require,
    package_data={},
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/vltr/sanic-toolbox/issues',
        'Source': 'https://github.com/vltr/sanic-toolbox',
    },
)
